// Copyright (C) 2011 - DIGITEO - Michael Baudin
//
// This file must be used under the terms of the GNU Lesser General Public License license :
// http://www.gnu.org/copyleft/lesser.html


function varargout = nisp_sobolsa ( varargin )
    // Compute sensitivity indices by Sobol, Ishigami, Homma.
    //
    // Calling Sequence
    //   s = nisp_sobolsa ( func , nx )
    //   s = nisp_sobolsa ( func , nx , randgen )
    //   s = nisp_sobolsa ( func , nx , randgen , n )
    //   [ s , nbevalf ] = nisp_sobolsa ( ... )
    //   [ s , nbevalf , st ] = nisp_sobolsa ( ... )
    //   [ s , nbevalf , st , mi , si ] = nisp_sobolsa ( ... )
    //
    // Parameters
    //   func : a function or a list, the name of the function to be evaluated. 
    //   nx : a 1-by-1 matrix of floating point integers, the number of inputs of the function.
    //   randgen : a function or a list, the random number generator. (default = uniform random variables)
    //   n : a 1-by-1 matrix of floating point integers (default n=10000), the number of Monte-Carlo experiments, for each sensitivity index
    //   s : a nx-by-1 matrix of doubles, the first order sensitivity indices
    //   nbevalf : a nx-by-1 matrix of doubles, the actual number of function evaluations.
    //   st : a nx-by-1 matrix of doubles, the total sensitivity indices
    //   mi : a m-by-nx matrix of doubles, the multi-indices of the variables in si, where m=2^nx - 1. Each row in mi represents a group of variables. We have mi(k,i) = 1 if Xi is in the group of variables and 0 if not. 
    //   si : a m-by-1 matrix of doubles, the partial sensitivity indices.
    //
    // Description
    // The algorithm uses the Sobol method to compute the first order 
    // sensitivity indices.
    //
    // This method assumes that all the input random variables are 
    // independent.
    //
    // Any optional input argument equal to the empty matrix will be set to its
    // default value.
    //
    // The function should have header y = func ( x ) where x is a m-by-nx matrix of doubles, 
    // where m is the number of experiments to perform, nx is the number of input 
    // random variables, and y is a m-by-1 matrix of doubles.
    //
    // It might happen that the function requires additionnal 
    // arguments to be evaluated.
    // In this case, we can use the following feature.
    // The function func can also be a list, with header y = f ( x , a1 , a2 , ... ).
    // In this case the func variable should hold the list (f,a1,a2,...) and the 
    // input arguments a1, a2, will be automatically be appended at the 
    // end of the calling sequence of rg.
    //
    // The random number generator must have header x = randgen ( m , i ) where 
    //   <itemizedlist>
    // <listitem>m is a 1-by-1 matrix of floating point integers, the number of 
    // experiments to perform</listitem>
    // <listitem>i is a 1-by-1 matrix of floating point integers representing the index of the 
    // input variable,</listitem>
    // <listitem>and x is a m-by-1 matrix of doubles. </listitem>
    //   </itemizedlist>
    // On output, x must contain random numbers sampled from the distribution 
    // function associated with the input variable #i, where i is in the set {1,2,...,nx}.
    //
    // Since the input random variables are independent, we can generate 
    // the samples associated with Xi independently from the samples for 
    // Xj, for i not equal to j.
    // This is why the callback randgen only needs the index i of the 
    // input random variable.
    // If there is some dependency between the inputs (e.g. correlation), 
    // then the randgen callback would require to generate all the samples 
    // dependently one from the other.
    // In this case, the current function cannot be used.
    //
    // It might happen that the random number generator requires additionnal 
    // arguments to be evaluated.
    // In this case, we can use the following feature.
    // The function randgen can also be a list, with header u = rg ( m , i , a1 , a2 , ... ).
    // In this case the randgen variable should hold the list (rg,a1,a2,...) and the 
    // input arguments a1, a2, will be automatically be appended at the 
    // end of the calling sequence of rg.
    // 
    // TODO : test it on a 4 input variables test case.
    //
    // Examples
    // // Compute the first order sensitivity indices of the ishigami function.
    // // Three random variables uniform in [-pi,pi].
    // function y = ishigami (x)
    //   a=7.
    //   b=0.1
    //   s1=sin(x(:,1))
    //   s2=sin(x(:,2))
    //   x34 = x(:,3).^4
    //   y(:,1) = s1 + a.*s2.^2 + b.*x34.*s1
    // endfunction
    // function x = myrandgen ( m , i )
    //   x = grand(m,1,"unf",-%pi,%pi)
    // endfunction
    // a=7.;
    // b=0.1;
    // exact.expectation = a/2;
    // exact.var = 1/2 + a^2/8 + b*%pi^4/5 + b^2*%pi^8/18;
    // exact.S1 = (1/2 + b*%pi^4/5+b^2*%pi^8/50)/exact.var;
    // exact.S2 = (a^2/8)/exact.var;
    // exact.S3 = 0;
    // n = 1000;
    // nx = 3;
    // [ s , nbevalf ] = nisp_sobolsa ( ishigami , nx , myrandgen , n );
    // mprintf("S1 : %f (expected = %f)\n", s(1), exact.S1);
    // mprintf("S2 : %f (expected = %f)\n", s(2), exact.S2);
    // mprintf("S3 : %f (expected = %f)\n", s(3), exact.S3);
    //
    // // See the variability of the sensitivity indices.
    // for k = 1 : 100
    //   [ s , nbevalf ] = nisp_sobolsa ( ishigami , nx , myrandgen , 1000 );
    //   sall(k,:) = s';
    // end
    // scf();
    // subplot(2,2,1);
    // histplot(10,sall(:,1));
    // xtitle("Variability of the sensitivity index for X1","S1","P(S1)");
    // subplot(2,2,2);
    // histplot(10,sall(:,2));
    // xtitle("Variability of the sensitivity index for X2","S2","P(S2)");
    // subplot(2,2,3);
    // histplot(10,sall(:,3));
    // xtitle("Variability of the sensitivity index for X3","S3","P(S3)");
    //
    // // See the convergence of the sensitivity indices
    // n=10;
    // stacksize("max");
    // for k = 1 : 100
    //   tic();  
    //   [ s , nbevalf ] = nisp_sobolsa ( ishigami , nx , myrandgen , n );
    //   sc(k,:) = s';
    //   t = toc();
    //   mprintf("Run #%d, n=%d, t=%.2f (s)\n",k,n,t);
    //   if ( t > 1 ) then
    //     break
    //   end
    //   n = 1.2*n;
    // end
    // h = scf();
    // subplot(1,2,1);
    // plot(1:k,sc(1:k,1),"bx-");
    // plot(1:k,sc(1:k,2),"ro-");
    // plot(1:k,sc(1:k,3),"g*-");
    // xtitle("Convergence of the sensitivity indices","Number of simulations","S");
    // legend(["S1","S2","S3"]);
    // subplot(1,2,2);
    // plot(1:k,abs(sc(1:k,1)-exact.S1),"bx-");
    // plot(1:k,abs(sc(1:k,2)-exact.S2),"ro-");
    // plot(1:k,abs(sc(1:k,3)-exact.S3),"g*-");
    // xtitle("Convergence of the sensitivity indices","Number of simulations","|S-exact|");
    // legend(["S1","S2","S3"]);
    // h.children(1).log_flags="lnn";
    //
    // // Compute the total sensivity indices.
    // [ s , nbevalf , st ] = nisp_sobolsa ( ishigami , nx , myrandgen )
    //
    // // Compute all the sensivity indices.
    // [ s , nbevalf , st , mi , si ] = nisp_sobolsa ( ishigami , nx , myrandgen )
    // // The ANOVA decomposition can be seen more easily in
    // [mi si]
    // // The sum of si is 1.
    // sum(si) // expected = 1.
    //
    // Authors
    //   Michael Baudin - 2011 - DIGITEO

    function s = sensitivityindex(ya,yc)
        // Returns the sensitivity index associated with experiments ya and yc.
        C = nisp_cov (ya, yc)
        s= C(1,2)/ (st_deviation(ya) * st_deviation(yc))
    endfunction

    function A = getsampling ( n , nx , __sobolfo_randgen__ )
        if ( typeof( __sobolfo_randgen__ ) == "list" ) then
            // List
            __sobolfo_randgen__f = __sobolfo_randgen__ (1);
        end
        A = zeros(n , nx)
        for i = 1 : nx
            if ( typeof( __sobolfo_randgen__ ) == "list" ) then
                // List
                Ai = __sobolfo_randgen__f ( n , i , __sobolfo_randgen__ (2:$));
            else
                // Macro or compiled macro
                Ai = __sobolfo_randgen__ ( n , i );
            end

            if ( size(Ai)<>[n 1] ) then
                errmsg = msprintf(gettext("%s: The size of the output of randgen is [%d,%d], which is different from [%d,%d]."), "nisp_sobolsa" , size(Ai,"r") , size(Ai,"c") , n , 1 );
                error(errmsg)
            end
            A(:,i) = Ai
        end
    endfunction

    function x = __nisp_sobol_rgdef__ ( m , i )
        x = grand(m,1,"def")
    endfunction

    function y = getfvalue ( x , __sobolfo_func__ )
        if ( typeof( __sobolfo_func__ ) == "list" ) then
            // List
            __sobolfo_func__f = __sobolfo_func__ (1);
            y = __sobolfo_func__f ( x , __sobolfo_func__ (2:$));
        else
            // Macro or compiled macro
            y = __sobolfo_func__ ( x );
        end
    endfunction

    function argin = argindefault ( rhs , vararglist , ivar , default )
        // Returns the value of the input argument #ivar.
        // If this argument was not provided, or was equal to the 
        // empty matrix, returns the default value.
        if ( rhs < ivar ) then
            argin = default
        else
            if ( vararglist(ivar) <> [] ) then
                argin = vararglist(ivar)
            else
                argin = default
            end
        end
    endfunction

    // Copyright (C) 2010 - DIGITEO - Michael Baudin
    //
    // This file must be used under the terms of the CeCILL.
    // This source file is licensed as described in the file COPYING, which
    // you should have received as part of this distribution.  The terms
    // are also available at
    // http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt

    function tf = number_ismember ( a , s )
        // Array elements that are members of set.
        //
        // Calling Sequence
        // tf = number_ismember ( a , s )
        //
        // Parameters
        // a : a ma-by-na matrix
        // s : a 1-by-ns matrix
        // tf : a ma-by-na matrix of booleans, tf(i,j) is %t if there is an entry in s matching a(i,j).
        //
        // Description
        // Search in a the entries which are in s.
        // 
        // Uses vectorized statements, based on the Kronecker product. 
        // The intermediate memory required is ma*na*ns.
        // 
        // Examples
        // a = (1:5)'
        // s = [0 2 4 6 8 10 12 14 16 18 20]
        // tf = number_ismember ( a , s )
        // expected = [%f %t %f %t %f]'
        //
        // // An example with a matrix
        // a = [
        // 7 35 14 86 76   
        // 15 51 24 96 49   
        // 35 40 46 35 93   
        // 85 34 74 82 22
        // ];
        // s = [51 74 22 15 86] 
        // tf = number_ismember(a, s)
        //
        // Authors
        // Copyright (C) 2010 - DIGITEO - Michael Baudin

        [lhs, rhs] = argn()
        apifun_checkrhs ( "number_ismember" , rhs , 2:2 )
        apifun_checklhs ( "number_ismember" , lhs , 0:1 )
        //
        // Check type
        apifun_checktype ( "number_ismember" , a , "a" , 1 , "constant" )
        apifun_checktype ( "number_ismember" , s , "s" , 2 , "constant" )
        //
        // Check size
        // a can have any shape
        if ( s==[] ) then
            tf=(ones(a)==zeros(a))
            return
        else
            apifun_checkvecrow ( "number_ismember" , s , "s" , 2 , size(s,"*") )
        end
        //
        // Check content
        // Nothing to do.
        //
        // Proceed...
        if ( a==[] ) then
            tf=[]
            return
        end
        //
        // Convert a into a column vector
        ma = size(a,"r")
        na = size(a,"c")
        a=a(:)
        //
        ns = size(s,"c")
        sp = s .*. ones(na*ma,1)
        ap = a .*. ones(1,ns)
        tf = or(ap==sp,"c")
        //
        // Convert tf into the same shape as a
        tf=matrix(tf,ma,na)
    endfunction



    [lhs, rhs] = argn();
    apifun_checkrhs ( "nisp_sobolsa" , rhs , 2 : 4 )
    apifun_checklhs ( "nisp_sobolsa" , lhs , [0 1 2 3 5] )
    //
    __sobolfo_func__ = varargin(1)
    nx = varargin(2)
    __sobolfo_randgen__ = argindefault ( rhs , varargin , 3 , __nisp_sobol_rgdef__ )
    n = argindefault ( rhs , varargin , 4 , 10000 )
    //
    // Check input arguments
    //
    // Check type
    apifun_checktype ( "nisp_sobolsa" , __sobolfo_func__ ,   "func" ,  1 , [ "function" "list" ] )
    apifun_checktype ( "nisp_sobolsa" , nx ,      "nx" ,     2 , "constant" )
    apifun_checktype ( "nisp_sobolsa" , __sobolfo_randgen__ ,  "randgen" , 3 , [ "function" "list" ] )
    apifun_checktype ( "nisp_sobolsa" , n ,  "n" , 4 , "constant" )
    //
    // Check size
    apifun_checkscalar ( "nisp_sobolsa" , nx ,       "nx" ,       2 )
    apifun_checkscalar ( "nisp_sobolsa" , n ,   "n" ,   4 )
    //
    // Check content
    apifun_checkgreq ( "nisp_sobolsa" , nx ,  "nx" ,  2 , 1 )
    apifun_checkgreq ( "nisp_sobolsa" , n ,   "n" ,   4 , 1 )
    //
    // Proceed...
    // Create a sampling A
    A = getsampling ( n , nx , __sobolfo_randgen__ )
    // Create a sampling B
    B = getsampling ( n , nx , __sobolfo_randgen__ )
    // Perform the experiments in A
    ya = getfvalue ( A , __sobolfo_func__ )
    // Compute the sensitivity indices
    for i = 1 : nx
        C(:,1:i-1) = B(:,1:i-1);
        C(:,i)=A(:,i);
        C(:,i+1:nx) = B(:,i+1:nx);
        yc = getfvalue ( C , __sobolfo_func__ )
        s(i) = sensitivityindex ( ya , yc );
    end
    nbevalf = n + nx * n
    //
    // Set outputs
    varargout(1) = s
    varargout(2) = nbevalf
    if ( lhs >= 3 ) then 
        //
        // If total sensitivity indices are to compute, compute them.
        for i = 1 : nx
            C(:,1:i-1) = A(:,1:i-1);
            C(:,i)=B(:,i);
            C(:,i+1:nx) = A(:,i+1:nx);
            yc = getfvalue ( C , __sobolfo_func__ )
            st(i) = 1 - sensitivityindex ( ya , yc );
        end
        nbevalf = nbevalf + nx * n
        varargout(2) = nbevalf
        varargout(3) = st
    end
    if ( lhs == 5 ) then 
        //
        // If partial sensitivity indices are to compute, compute them.
        //
        // Compute all the combinations of variables Xi1, Xi2, ..., Xip.
        // If nx=3, the matrix is :
        // 0 0 0
        // 1 0 0
        // 0 1 0
        // 0 0 1
        // 1 1 0
        // 1 0 1
        // 0 1 1
        // 1 1 1
        // It represents the possible combinations of variables: 
        // 1 means that Xi is in the subset, 0 means no variable Xi.
        mi = specfun_combinerepeat ( [1 0] , nx )'
        ks = sum(mi,"c")
        // Sort by increasing number of variables.
        [ignored,korder]=gsort(ks,"g","i")
        mi = mi(korder,:)
        // Eliminate the combination of variables 
        // where there are less than 1 variables, i.e. delete 
        // the row [0 0 0 ... 0].
        mi(1,:)=[]
        nrows = size(mi,"r")
        si=zeros(nrows,1)
        // The first nx sensitivity indices are already 
        // computed: just store them.
        for k = 1 : nx
            // Get the variable i in this row: 
            // there is a unique non-zero in the row k.
            i = find(mi(k,:)==1)
            si(k) = s(i)
        end
        // Compute the sensitivity indices for all these combinations
        // of at least two variables.
        for k = nx + 1 : nrows
            // Compute the sensitivity index for the group of 
            // variables in the row mi(k,:).
            knovar = find(mi(k,:)==0);
            C(:,knovar) = B(:,knovar);
            kvar = find(mi(k,:)==1)
            C(:,kvar)=A(:,kvar);
            yc = getfvalue ( C , __sobolfo_func__ );
            si(k) = sensitivityindex(ya,yc);
            // Remove the lower order sensitivity indices.
            for i = 1 : k-1
                ivar = find(mi(i,:)==1);
                // The sensitivity index si(i) must be subtracted 
                // from si(k) is all variables in ivar are in kvar.
                tf = number_ismember ( ivar , kvar );
                if ( and(tf) ) then
                    si(k) = si(k) - si(i);
                end
            end
        end
        nbevalf = nbevalf + ( nrows - nx -1 ) * n
        varargout(2) = nbevalf
        varargout(4) = mi
        varargout(5) = si
    end
endfunction

